/**
 * 
 * Compilation with nvc or pgcc
 * ----------------------------
 * nvc -acc -ta=tesla -Minfo -fast -o mandelbrot 02_mandelbrot.c
 * pgcc -acc -ta=tesla -Minfo -fast -o mandelbrot 02_mandelbrot.c
 *
 * Compilation with gcc
 * -------------------
 * Leads to internal compiler error :(
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#define  WIDTH 8192
#define HEIGHT 8192

#define MAX_ITERS 100
#define MAX_COLOR 255

#define XMIN -1.7
#define XMAX  0.5
#define YMIN -1.2
#define YMAX  1.2

#define DX ((XMAX - XMIN) / WIDTH)
#define DY ((YMAX - YMIN) / HEIGHT)

#pragma acc routine seq
unsigned char mandelbrot(int px, int py) {
  const double x0 = XMIN + px * DX;
  const double y0 = YMIN + py * DY;
  
  double x = 0.0;
  double y = 0.0;
  
  int i;
  for(i = 0; (x * x + y * y) < 4.0 && i < MAX_ITERS; i++) {
    double xtemp = x * x - y * y + x0;
    
    y = 2 * x * y + y0;
    x = xtemp;
  }

  return (double)MAX_COLOR*i/MAX_ITERS;
}

int main(int argc, char* argv[]) {
  struct timeval start, end;

  size_t bytes = WIDTH * HEIGHT * sizeof(unsigned char);
  unsigned char* image=(unsigned char*)malloc(bytes);
  
  int num_blocks = 8;
  int block_size = (HEIGHT / num_blocks) * WIDTH;
  
  FILE* fp = fopen("image.pgm", "wb");
  fprintf(fp, "P5\n%s\n%d %d\n%d\n","#comment", WIDTH, HEIGHT, MAX_COLOR);
  
  gettimeofday(&start, NULL);

  #pragma acc data create(image[WIDTH*HEIGHT])
  for(int block = 0; block < num_blocks; ++block) {
    int start = block * (HEIGHT/num_blocks);
    int end   = start + (HEIGHT/num_blocks);
    
    #pragma acc parallel loop async(block%2)
    for (int y = start; y < end; y++) {
      for (int x = 0; x < WIDTH; x++) {
        image[x + y * WIDTH] = mandelbrot(x, y);
      }
    }

    #pragma acc update host(image[block*block_size:block_size]) async(block%2)
  }

  #pragma acc wait

  gettimeofday(&end, NULL);
  double elapsed = (double)(end.tv_sec - start.tv_sec) + (end.tv_usec-start.tv_usec)*1E-6;
  printf("Time: %lf seconds.\n", elapsed);
  
  fwrite(image, sizeof(unsigned char), WIDTH * HEIGHT, fp);
  fclose(fp);

  free(image);
  
  return 0;
}