/**
 * 
 * Compilation with clang
 * ----------------------
 * nvc -fopenmp -fopenmp-targets=nvptx64 -O2 -o mandelbrot 02_mandelbrot.c
 *
 * Compilation with gcc
 * -------------------
 * gcc -fopenmp -foffload=nvptx-none -Ofast -o mandelbrot 02_mandelbrot.c
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

#define  WIDTH 8192
#define HEIGHT 8192

#define MAX_ITERS 100
#define MAX_COLOR 255

#define XMIN -1.7
#define XMAX  0.5
#define YMIN -1.2
#define YMAX  1.2

#define DX ((XMAX - XMIN) / WIDTH)
#define DY ((YMAX - YMIN) / HEIGHT)

#pragma omp declare target
unsigned char mandelbrot(int px, int py);
#pragma omp end declare target

unsigned char mandelbrot(int px, int py) {
  const double x0 = XMIN + px * DX;
  const double y0 = YMIN + py * DY;
  
  double x = 0.0;
  double y = 0.0;
  
  int i;
  for(i = 0; (x * x + y * y) < 4.0 && i < MAX_ITERS; i++) {
    double xtemp = x * x - y * y + x0;
    
    y = 2 * x * y + y0;
    x = xtemp;
  }

  return (double)MAX_COLOR*i/MAX_ITERS;
}

int main(int argc, char* argv[]) {
  size_t bytes = WIDTH * HEIGHT * sizeof(unsigned char);
  unsigned char* image=(unsigned char*)malloc(bytes);
  
  int num_blocks = 8;
  int block_size = (HEIGHT / num_blocks) * WIDTH;
  
  FILE* fp = fopen("image.pgm", "wb");
  fprintf(fp, "P5\n%s\n%d %d\n%d\n","#comment", WIDTH, HEIGHT, MAX_COLOR);
  
  double start = omp_get_wtime();

  #pragma omp target data map(alloc: image[0:WIDTH*HEIGHT])
  for(int block = 0; block < num_blocks; ++block) {
    int start = block * (HEIGHT/num_blocks);
    int end   = start + (HEIGHT/num_blocks);
    
    #pragma omp target teams distribute parallel for collapse(2) \
      depend(inout:image[block*block_size]) nowait
    for (int y = start; y < end; y++) {
      for (int x = 0; x < WIDTH; x++) {
        image[x + y * WIDTH] = mandelbrot(x, y);
      }
    }
    
    #pragma omp target update from(image[block*block_size:block_size]) \
      depend(inout:image[block*block_size]) nowait
  }
  
  #pragma omp taskwait

  double elapsed = omp_get_wtime() - start;
  printf("Time: %lf seconds.\n", elapsed);
  
  fwrite(image, sizeof(unsigned char), WIDTH * HEIGHT, fp);
  fclose(fp);

  free(image);
  
  return 0;
}